from unv.app.core import create_component_settings


DEFAULT = {
    'hosts': {},
    'components': {},
}

SCHEMA = {
    'hosts': {
        'type': 'dict',
        'keyschema': {'type': 'string'},
        'valueschema': {
            'type': 'dict',
            'schema': {
                'public': {'type': 'string'},
                'private': {'type': 'string'},
                'components': {'type': 'list', 'schema': {'type': 'string'}}
            }
        }
    },
    'components': {'allow_unknown': True}
}

SETTINGS = create_component_settings('deploy', DEFAULT, SCHEMA)



from pathlib import Path

from watchgod import awatch

from ...tasks import DeployComponentTasksBase, local, register
from ...helpers import ComponentSettingsBase, get_hosts

from ..python import PythonComponentTasks, PythonComponentSettings
from ..systemd import SystemdTasksMixin


class AppComponentSettings(ComponentSettingsBase):
    NAME = 'app'
    DEFAULT = {
        'bin': 'app',
        'instance': 1,
        'settings': 'secure.production',
        'systemd': {
            'template': 'app.service',
            'name': 'app_{instance}.service',
            'boot': True,
            'instances': {'count': 1},
            'context': {
                'limit_nofile': 2000,
                'description': "Application description",
            }
        },
        'watch': {
            'dir': './src',
            'exclude': ['__pycache__']
        }
    }

    @property
    def python(self):
        settings = self._data.get('python', {})
        settings['user'] = self.user
        return PythonComponentSettings(settings)

    @property
    def bin(self):
        return str(self.python.root_abs / 'bin' / self._data['bin'])

    @property
    def module(self):
        return self._data['settings']

    @property
    def instance(self):
        return self._data['instance']

    @property
    def watch_dir(self):
        return Path(self._data['watch']['dir'])

    @property
    def watch_exclude(self):
        return self._data['watch']['exclude']


class AppComponentTasks(DeployComponentTasksBase, SystemdTasksMixin):
    SETTINGS = AppComponentSettings()
    NAMESPACE = 'app'

    def __init__(self, user, host, settings=None):
        super().__init__(user, host, settings)
        self._python = PythonComponentTasks(user, host, self._settings.python)

    @register
    @local
    async def watch(self):
        directory = self._settings.watch_dir
        site_packages_abs = self._settings.python.site_packages_abs

        async for _ in awatch(directory):
            for _, host in get_hosts(self.NAMESPACE):
                with self._set_user(self._settings.user), self._set_host(host):
                    for sub_dir in directory.iterdir():
                        if not sub_dir.is_dir():
                            continue
                        await self._rsync(
                            sub_dir, site_packages_abs / sub_dir.name,
                            self._settings.watch_exclude
                        )
                    await self.restart()

    @register
    async def build(self):
        await self._create_user()
        await self._python.build()

    @register
    async def shell(self):
        return await self._python.shell()

    @register
    async def sync(self):
        name = (await self._local('python setup.py --name')).strip()
        version = (await self._local('python setup.py --version')).strip()
        package = f'{name}-{version}.tar.gz'

        await self._local('pip install -e .')
        await self._local('python setup.py sdist bdist_wheel')
        await self._upload(Path('dist', package))
        await self._local('rm -rf ./build ./dist')
        await self._python.pip(f'install -I {package}')
        await self._rmrf(Path(package))
        await self._upload(Path('secure'))
        await self._sync_systemd_units()



import asyncio
import logging
import contextlib

from pathlib import Path

import jinja2

from unv.utils.tasks import TasksBase, TasksManager, TaskRunError, register

from .helpers import get_hosts, as_root
from .settings import SETTINGS
from .helpers import ComponentSettingsBase


def parallel(task):
    task.__parallel__ = True
    return task


def local(task):
    task.__local__ = True
    return task


class DeployTasksBase(TasksBase):
    def __init__(self, user, host):
        self._user = user
        self._public_ip = host['public']
        self._private_ip = host['private']
        self._port = host.get('ssh', 22)
        self._current_prefix = ''
        self._logger = logging.getLogger(self.__class__.__name__)

    @contextlib.contextmanager
    def _set_user(self, user):
        old_user = self._user
        self._user = user
        yield self
        self._user = old_user

    @contextlib.contextmanager
    def _set_host(self, host):
        old_public_ip, old_private_ip, old_port =\
            self._public_ip, self._private_ip, self._port
        self._public_ip, self._private_ip, self._port =\
            host['public'], host['private'], host.get('ssh', 22)
        yield self
        self._public_ip, self._private_ip, self._port =\
            old_public_ip, old_private_ip, old_port

    @contextlib.contextmanager
    def _prefix(self, command):
        old_prefix = self._current_prefix
        self._current_prefix = f'{self._current_prefix} {command} '
        yield
        self._current_prefix = old_prefix

    @contextlib.asynccontextmanager
    async def _cd(self, path: Path, temporary=False):
        if temporary:
            await self._mkdir(path, delete=True)
        with self._prefix(f'cd {path} &&'):
            yield
        if temporary:
            await self._rmrf(path)

    @as_root
    async def _sudo(self, command, strip=True):
        """Run command on server as root user."""
        return await self._run(command, strip)

    async def _create_user(self):
        """Create user if not exist and sync ssh keys."""
        user = self._user

        with self._set_user('root'):
            try:
                await self._run("id -u {}".format(user))
            except TaskRunError:
                await self._run(
                    "adduser --quiet --disabled-password"
                    " --gecos \"{0}\" {0}".format(user)
                )

                local_ssh_public_key = Path('~/.ssh/id_rsa.pub')
                local_ssh_public_key = local_ssh_public_key.expanduser()
                keys_path = Path(
                    '/', 'home' if user != 'root' else '',
                    user, '.ssh'
                )

                await self._mkdir(keys_path)
                await self._run(f'chown -hR {user} {keys_path}')
                await self._run('echo "{}" >> {}'.format(
                    local_ssh_public_key.read_text().strip(),
                    keys_path / 'authorized_keys'
                ))

    @as_root
    async def _apt_install(self, *packages):
        with self._prefix('DEBIAN_FRONTEND=noninteractive'):
            await self._run('apt-get update -y -q')
            await self._run('apt-get upgrade -y -q')
            await self._run(
                'apt-get install -y -q --no-install-recommends '
                '--no-install-suggests {}'.format(' '.join(packages))
            )

    async def _run(self, command, strip=True, interactive=False) -> str:
        self._logger.debug(
            f'run [{self._user}@{self._public_ip}:{self._port}] '
            f'{self._current_prefix}{command}'
        )
        interactive_flag = '-t' if interactive else ''
        response = await self._local(
            f"ssh {interactive_flag} -p {self._port} "
            f"{self._user}@{self._public_ip} "
            f"'{self._current_prefix}{command}'",
            interactive=interactive
        ) or ''
        if strip:
            response = response.strip()
        return response

    async def _rmrf(self, path: Path):
        await self._run(f'rm -rf {path}')

    async def _mkdir(self, path: Path, delete=False):
        if delete:
            await self._rmrf(path)
        await self._run(f'mkdir -p {path}')

    async def _upload(self, local_path: Path, path: Path = '~/'):
        await self._local(
            f'scp -r -P {self._port} {local_path} '
            f'{self._user}@{self._public_ip}:{path}'
        )

    async def _rsync(self, local_dir, root_dir, exclude=None):
        exclude = [f"--exclude '{path}'" for path in exclude or []]
        exclude = ' '.join(exclude)
        await self._local(
            f'rsync -rave ssh --delete {exclude} {local_dir}/ '
            f'{self._user}@{self._public_ip}:{root_dir}'
        )

    async def _upload_template(
            self, local_path: Path, path: Path, context: dict = None):
        context = context or {}
        render_path = Path(f'{local_path}.render')
        template = jinja2.Template(local_path.read_text())
        render_path.write_text(template.render(context))
        try:
            await self._upload(render_path, path)
        finally:
            render_path.unlink()

    async def _download_and_unpack(self, url: str, dest_dir: Path = Path('.')):
        await self._run(f'wget -q {url}')
        archive = url.split('/')[-1]
        await self._run(f'tar xf {archive}')
        archive_dir = archive.split('.tar')[0]

        await self._mkdir(dest_dir)
        await self._run(f'mv {archive_dir}/* {dest_dir}')

        await self._rmrf(archive)
        await self._rmrf(archive_dir)

    @register
    async def ssh(self):
        return await self._run('bash', interactive=True)


class DeployComponentTasksBase(DeployTasksBase):
    SETTINGS = None

    def __init__(self, user, host, settings=None):
        super().__init__(user, host)
        settings = settings or self.__class__.SETTINGS

        if settings is None or not isinstance(settings, ComponentSettingsBase):
            raise ValueError(
                "Provide correct 'SETTINGS' value "
                "shoult be an instance of class 'ComponentSettingsBase' not "
                f"[{settings}] value and type {type(settings)}"
            )

        self._settings = settings


class DeployTasksManager(TasksManager):
    def run_task(self, task_class, name, args):
        method = getattr(task_class, name)
        is_local = hasattr(method, '__local__')
        is_parallel = hasattr(method, '__parallel__')

        if issubclass(task_class, DeployTasksBase):
            if is_local:
                user = 'local'
                hosts = [{'public': None, 'private': None, 'ssh': 0}]
            else:
                user, hosts = self._select_hosts(task_class.NAMESPACE)

            tasks = [
                getattr(task_class(user, host), name)(*args)
                for host in hosts
            ]

            if is_parallel:
                async def run():
                    await asyncio.gather(*tasks)
                asyncio.run(run())
            else:
                for task in tasks:
                    asyncio.run(task)
        else:
            return super().run_task(task_class, name, args)

    def _select_hosts(self, name: str = ''):
        return (
            SETTINGS['components'].get(name, {}).get('user', name),
            [host for _, host in get_hosts(name)]
        )



import functools
import copy
import inspect

from pathlib import Path

from unv.utils.collections import update_dict_recur

from .settings import SETTINGS


class ComponentSettingsBase:
    def __init__(self, settings=None, root=None):
        if settings is None:
            settings = SETTINGS['components'].get(self.__class__.NAME, {})
        # TODO: add schema validation
        self._data = update_dict_recur(
            copy.deepcopy(self.__class__.DEFAULT), settings)
        self.local_root = root or Path(inspect.getfile(self.__class__)).parent

    @property
    def user(self):
        return self._data.get('user', self.__class__.NAME)

    @property
    def enabled(self):
        if 'enabled' in self._data:
            return self._data['enabled']
        for _, host in get_hosts():
            if self.__class__.NAME in host['components']:
                return True

    @property
    def home(self):
        return Path('~')

    @property
    def home_abs(self):
        return Path('/', 'home', self.user)

    @property
    def systemd(self):
        return self._data.get('systemd', {})

    @property
    def root(self):
        return self.home / self._data['root']

    @property
    def root_abs(self):
        return self.home_abs / self._data['root']


def as_user(user, func=None):
    """Task will run from any user, sets to env.user."""

    def decorator(func):
        @functools.wraps(func)
        async def wrapper(self, *args, **kwargs):
            with self._set_user(user):
                return await func(self, *args, **kwargs)
        return wrapper

    return decorator if func is None else decorator(func)


def as_root(func):
    return as_user('root', func)


def get_hosts(component=''):
    for key, value in SETTINGS['hosts'].items():
        if component in value.get('components', []) or not component:
            yield key, value
